@@ -24,7 +24,7 @@
 #include <stdarg.h>
 #include "ass_types.h"
 
-#define LIBASS_VERSION 0x01700000
+#define LIBASS_VERSION 0x01703000
 
 #ifdef __cplusplus
 extern "C" {
@@ -324,9 +324,14 @@ void ass_set_extract_fonts(ASS_Library *priv, int extract);
  *   ScaledBorderAndShadow=yes
  *
  * \param priv library handle
- * \param list NULL-terminated list of strings
+ * \param list NULL-terminated list of strings,
+ *             copied and never modified by libass
  */
+#ifdef __cplusplus
+void ass_set_style_overrides(ASS_Library *priv, const char *const *list);
+#else
 void ass_set_style_overrides(ASS_Library *priv, char **list);
+#endif
 
 /**
  * \brief Explicitly process style overrides for a track.
@@ -690,7 +695,7 @@ void ass_free_event(ASS_Track *track, int eid);
  * \param data string to parse
  * \param size length of data
  */
-void ass_process_data(ASS_Track *track, char *data, int size);
+void ass_process_data(ASS_Track *track, const char *data, int size);
 
 /**
  * \brief Parse Codec Private section of the subtitle stream, in Matroska
@@ -699,7 +704,7 @@ void ass_process_data(ASS_Track *track, char *data, int size);
  * \param data string to parse
  * \param size length of data
  */
-void ass_process_codec_private(ASS_Track *track, char *data, int size);
+void ass_process_codec_private(ASS_Track *track, const char *data, int size);
 
 /**
  * \brief Parse a chunk of subtitle stream data. A chunk contains exactly one
@@ -715,7 +720,7 @@ void ass_process_codec_private(ASS_Track *track, char *data, int size);
  * \param timecode starting time of the event (milliseconds)
  * \param duration duration of the event (milliseconds)
  */
-void ass_process_chunk(ASS_Track *track, char *data, int size,
+void ass_process_chunk(ASS_Track *track, const char *data, int size,
                        long long timecode, long long duration);
 
 /**
@@ -747,8 +752,8 @@ void ass_flush_events(ASS_Track *track);
  * if both versions are valid and exist.
  * On all other systems there is no need for special considerations like that.
 */
-ASS_Track *ass_read_file(ASS_Library *library, char *fname,
-                         char *codepage);
+ASS_Track *ass_read_file(ASS_Library *library, const char *fname,
+                         const char *codepage);
 
 /**
  * \brief Read subtitles from memory.
@@ -759,7 +764,7 @@ ASS_Track *ass_read_file(ASS_Library *library, char *fname,
  * \return newly allocated track or NULL on failure
 */
 ASS_Track *ass_read_memory(ASS_Library *library, char *buf,
-                           size_t bufsize, char *codepage);
+                           size_t bufsize, const char *codepage);
 /**
  * \brief Read styles from file into already initialized track.
  * \param fname file name
@@ -771,7 +776,7 @@ ASS_Track *ass_read_memory(ASS_Library *library, char *buf,
  * if both versions are valid and exist.
  * On all other systems there is no need for special considerations like that.
  */
-int ass_read_styles(ASS_Track *track, char *fname, char *codepage);
+int ass_read_styles(ASS_Track *track, const char *fname, const char *codepage);
 
 /**
  * \brief Add a memory font.
@@ -802,6 +807,22 @@ void ass_clear_fonts(ASS_Library *library);
  */
 long long ass_step_sub(ASS_Track *track, long long now, int movement);
 
+/**
+ * \brief Allocates memory that can be safely freed by libass later.
+ * Use this to allocate buffers you'll use to manually modify ASS_Track events
+ * or related structures.
+ * \param size number of bytes to allocate
+ * \return pointer to the allocated buffer, or NULL on failure
+ */
+void *ass_malloc(size_t size);
+
+/**
+ * \brief Releases memory previously allocated within libass.
+ * Use this to free memory allocated using ass_malloc.
+ * \param ptr pointer to the buffer to release; passing NULL is a no-op
+ */
+void ass_free(void *ptr);
+
 #undef ASS_DEPRECATED
 #undef ASS_DEPRECATED_ENUM
 
