@@ -16,11 +16,12 @@
  * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
  */
 
-#ifndef INTEL_CPUID_H
-#define INTEL_CPUID_H
+#ifndef LIBASS_X86_CPUID_H
+#define LIBASS_X86_CPUID_H
 
 #include <stdint.h>
 
+uint32_t ass_has_cpuid( void );
 void ass_get_cpuid( uint32_t *eax, uint32_t *ebx, uint32_t *ecx, uint32_t *edx);
 void ass_get_xgetbv( uint32_t op, uint32_t *eax, uint32_t *edx );
 
